# Supporting R script for: Brierley, L., Pedersen, A.B.P., and Woolhouse, M.E.J. Tissue Tropism and Transmission Ecology Predict Virulence of Human RNA Viruses.
# Prepared by L Brierley, 4/12/18

#############################
# Data and analytical setup #
#############################

# Set random seed for reproducible analysis
set.seed(1353)

# Read in required packages
library(rpart)				    
library(rpart.plot)	
library(RColorBrewer)
library(caret)
library(ROCR)
library(randomForest)
library(missForest)
library(pdp)

# Read in and process data, selecting only those viruses with reliable virulence data for includsion in analysis
vir <- read.csv("Virulence_data_Brierley_et_al_30_11_18.csv", sep=",", header=T)
vir <- subset(vir, Included==1)

# Remove 'Unassigned' as a family and rearrange by genome type
vir$Family <- droplevels(vir$Family)
vir$Family <- factor(vir$Family, levels=c("Arenaviridae","Bornaviridae","Filoviridae","Hantaviridae","Nairoviridae","Orthomyxoviridae","Paramyxoviridae","Peribunyaviridae","Phenuiviridae","Pneumoviridae","Rhabdoviridae","Astroviridae","Caliciviridae","Coronaviridae","Flaviviridae","Hepeviridae","Picornaviridae","Togaviridae","Picobirnaviridae","Reoviridae","Retroviridae"), labels=c("Arena-","Borna-","Filo-","Hanta-","Nairo-","Orthomyxo-","Paramyxo-","Peribunya-","Phenui-","Pneumo-","Rhabdo-","Astro-","Calici-","Corona-","Flavi-","Hepe-","Picorna-","Toga-","Picobirna-","Reo-","Retro-"))

# Convert 0/1 columns to factors
factor_columns <- c(names(vir[,grepl( "^Tr\\.|^Fatal\\.|^Tp\\.|^Host\\." , names(vir))]))
vir[factor_columns] <- lapply(vir[factor_columns], factor)

# Establish test and training set based on severity and taxonomy (if only a single virus exists for a Virulence-Genus combination, it defaults to the training set)
training_row_nos = createDataPartition(paste(vir$Virulence, vir$Genus), p = 0.75, list = FALSE)
vir_train <- vir[training_row_nos,]
vir_test <- vir[-training_row_nos,]
# Store test set virus species names
vir_test_names <- vir_test$Species

# Create True Skill Statistic function
trueskillstat <- function(matrix){
  as.numeric(matrix$byClass[1] + matrix$byClass[2] - 1)
}

# Specify predictor variables to use in models as a formula
formula_used <- formula(Virulence ~ Family+Genome+
                            Tr.level+
                            Tr.primary+
                            Tr.direct.contact+Tr.faecal.oral+Tr.respiratory+Tr.vector+Tr.multiple+Tr.food.borne+Tr.vertical+
                            Tp.primary+
                            Tp.vascular+Tp.circulatory+Tp.gastrointestinal+Tp.hepatic+Tp.neural+Tp.respiratory+Tp.cardiac+Tp.joints+Tp.renal+Tp.reproductive+Tp.sensory+Tp.skin+Tp.muscular+Tp.endocrine+Tp.multiple+
                            Host.range+Host.human.only+Host.nh.primates+Host.other.mammal+Host.bird)

################################
# Classification tree analysis #
################################

# Construct and store optimal branching size for 1000 repeats of a single classification tree
nsplitopt <- vector(mode = "integer", length = 1000)
for (i in 1:length(nsplitopt)) {
  cp <- rpart(formula_used, data=vir_train, minsplit=10)$cptable
  nsplitopt[i] <- cp[which.min(cp[,"xerror"]), "nsplit"]
}

# Extract consensus branching size
consensus_size <- unique(nsplitopt)[which.max(tabulate(match(nsplitopt, unique(nsplitopt))))]

# Build single classification stree and store complexity parameter table
class_tree <- rpart(formula_used, data=vir_train, method="class", minsplit=10)
cptabledf <- as.data.frame(class_tree$cptable)

# Prune single tree to the consensus branching size
class_tree_pruned <- prune(class_tree, cp = cptabledf$CP[which(cptabledf$nsplit==consensus_size)])

# Visualise the tree
prp(class_tree_pruned, extra=2, type=3, faclen=2, varlen=0, tweak=1.07, fallen.leaves=TRUE, ycompress=FALSE, yesno=F, box.col=c("gray80", "gray50")[class_tree_pruned$frame$yval])

# Prediction diagnostics and accuracy statistics for training set
predict_tree_class_train <- predict(class_tree_pruned, type="class")
print(matrix_tree_train <- confusionMatrix(predict_tree_class_train, vir_train$Virulence, positive="severe"))

# Prediction diagnostics and accuracy statistics for test set
predict_tree_class_test <- predict(class_tree_pruned, newdata=vir_test, type="class")
print(matrix_tree_test <- confusionMatrix(predict_tree_class_test, vir_test$Virulence, positive="severe")) 

# View predictions from the classification tree for individual virus species in test set
data.frame(Species = as.character(vir_test$Species), Virulence = as.character(vir_test$Virulence), Predicted = as.character(predict_tree_class_test))

# Calculate ROC curve for classification tree
predict_tree_prob_test <- predict(class_tree_pruned, newdata=vir_test, type="prob")
roc_pred_tree <- prediction(predict_tree_prob_test[,2], as.numeric(vir_test$Virulence)-1)
# Calculate AUROC for classification tree
print(AUC_class_tree <- performance(roc_pred_tree, measure="auc")@y.values)
# Calculate TSS for classification tree
trueskillstat(matrix_tree_test)

##########################
# Random forest analysis #
##########################

# Impute missing data, itself from a random forest
data_imp_train <- missForest(vir_train[, all.vars(formula_used)])
vir_train <- data_imp_train$ximp
data_imp_test <- missForest(vir_test[, all.vars(formula_used)])
vir_test <- data_imp_test$ximp

# Build random forest model from 5000 individual bootstrapped trees
random_forest <- randomForest(formula_used, 
                             dat= vir_train,
                             importance = TRUE, 
                             na.action = na.exclude, 
                             keep.forest = TRUE, 
                             ntree = 5000, 
                             proximity = TRUE,
                             nodesize = 5)

# Visaualise convergence of error rates for severe, nonsevere, and overall
plot(random_forest)

# Extract variable importances and calculate relative mean decrease in Gini coefficient associated with each predictor
varimp <- as.data.frame(random_forest$importance)
varimp$relGini <- varimp$MeanDecreaseGini/max(varimp$MeanDecreaseGini)
# Append variable names to data frame
varimp$name <- rownames(varimp)
rownames(varimp) <- NULL

# Extract partial dependences as marginal relative change in log-odds of severe disease, and as marginal probability of severe disease
list_PD <- list_prob <<- vector("list", length(all.vars(formula_used))-1) 
names(list_PD) <- names(list_prob) <- all.vars(formula_used)[2:length(all.vars(formula_used))]

for (i in 2:length(all.vars(formula_used))){
  list_PD[[i-1]] <- partial(random_forest, pred.var=all.vars(formula_used)[i], which.class="severe")
}

for (i in 2:length(all.vars(formula_used))){
  list_prob[[i-1]] <- partial(random_forest, pred.var=all.vars(formula_used)[i], which.class="severe", prob=TRUE)
}

# Arrange partial dependences into data frame
df_PD_prob <- data.frame(Trait = unlist(data.frame(t(sapply(list_prob,c)))$Family),
                         PD = round(unlist(data.frame(t(sapply(list_PD,c)))$yhat),3),
                         Prob = round(unlist(data.frame(t(sapply(list_prob,c)))$yhat),3))
df_PD_prob <- data.frame(Variable = gsub('[[:digit:]]+', '', rownames(df_PD_prob)), df_PD_prob)
rownames(df_PD_prob) <- NULL

# Prediction diagnostics and accuracy statistics for training set
predict_forest_class_train <- predict(random_forest, type="class")
print(matrix_forest_train <- confusionMatrix(predict_forest_class_train, vir_train$Virulence, positive="severe")) 

# Prediction diagnostics and accuracy statistics for test set
predict_forest_class_test <- predict(random_forest, newdata=vir_test, type="response")
print(matrix_forest_test <- confusionMatrix(predict_forest_class_test, vir_test$Virulence, positive="severe")) 

# View predictions from the random forest for individual virus species in test set, as predicted category and as probability of severe disease
predict_forest_prob_test <- predict(random_forest, newdata=vir_test, type="prob")
data.frame(Species = as.character(vir_test_names), Virulence = as.character(vir_test$Virulence), Predicted = as.character(predict_forest_class_test), Probability = predict_forest_prob_test[,2])

# Calculate ROC curve for random forest
roc_pred_forest <- prediction(predict_forest_prob_test[,2], as.numeric(vir_test$Virulence)-1)
# Calculate AUROC for random forest
print(AUC_class_forest <- performance(roc_pred_forest, measure="auc")@y.values)
# Calculate TSS for random forest
trueskillstat(matrix_forest_test)

# Plot ROC curves
plot(performance(roc_pred_tree, measure="tpr", x.measure="fpr"), lwd=2.5, cex.axis=2)
plot(performance(roc_pred_forest, measure="tpr", x.measure="fpr"), add=TRUE, lwd=2.5, col = "red")
abline(a=0, b= 1)
legend(0.65,0.2, 
       c("Classification tree","Random forest"),
       lty=c(1,1),
       lwd=c(2.5,2.5),col=c("black","red")) 